from dataclasses import dataclass
from datetime import datetime
from typing import Iterator

from .base import BaseAPIModel


__all__ = ['Population', 'District']


@dataclass
class District:
    """Wrapper class for district data
    
    Attributes
    ----------
    name : str
        the name of the district
        
    population : int
        the population of the district
        
    last_updated : datetime
        when the district's population was last updated by the server
    """
    name: str
    population: int
    last_updated: datetime


class Population(BaseAPIModel):
    """"Wrapper class for /population response

    A tuple-like class containing `District` objects, sorted by population

    Attributes
    ----------
    total : int
        the total population of Toontown Rewritten

    last_updated : datetime
        the time of the last population update
    """

    __slots__ = ['total', 'last_updated']

    def __init__(self, **payload) -> None:
        self.total: int = payload.get('totalPopulation')
        self.last_updated = last_updated = datetime.fromtimestamp(payload.get('lastUpdated'))

        iterable = tuple(sorted([
            District(name, population, last_updated)
            for name, population in payload.get('populationByDistrict').items()
        ], key=lambda district: district.population))
        
        super().__init__(iterable)

    def __getitem__(self, index: int) -> District:
        return self._iterable.__getitem__(index)

    def __iter__(self) -> Iterator[District]:
        return self._iterable.__iter__()

    def __next__(self) -> District:
        return next(self._iterable)



from abc import abstractmethod, ABC
from pathlib import Path
from typing import Optional, Union

from .models import *
from .httpclient import BaseHTTPClient, SyncHTTPClient, AsyncHTTPClient, Route


class BaseToontownClient(ABC):
    @abstractmethod
    def __init__(self, httpclient: BaseHTTPClient) -> None:
        self.http = httpclient

    @abstractmethod
    def connect(self) -> None:
        """Connect to the HTTP client
        
        Must be called before using any other methods in this class
        """

    @abstractmethod
    def close(self) -> None:
        """Closes connection to the HTTP client"""

    @abstractmethod
    def doodles(self) -> Doodles:
        """Request Doodle data from the Toontown Rewritten API"""

    @abstractmethod
    def field_offices(self) -> FieldOffices:
        """Request Field Office data from the Toontown Rewritten API"""

    @abstractmethod
    def invasions(self) -> Invasions: 
        """Request Invasion data from the Toontown Rewritten API"""

    @abstractmethod
    def login(
        self, 
        *, 
        username: Optional[str] = None, 
        password: Optional[str] = None,
        response_token: Optional[str] = None,
        queue_token: Optional[str] = None,
    ) -> Login: 
        """Request to log into Toontown Rewritten's game server

        Must provide a username and password, a response token, or a queue token
        
        Parameters
        ----------
        username : Optional[str] = None
            optional username parameter, must also provide password if given

        password : Optional[str] = None
            optional password parameter, must also provide username if given

        response_token : Optional[str] = None
            optional response token parameter, obtained after initial login request with username and password 
            if you are required to authenticate with ToonGuard

        queue_token : Optional[str] = None
            optional queue token parameter, obtained after intial login request with username and password or 
            response token if you are required to wait in the login queue
        """

    @abstractmethod
    def population(self) -> Population: 
        """Request Population data from the Toontown Rewritten API"""

    @abstractmethod
    def silly_meter(self) -> SillyMeter: 
        """Request Silly Meter data from the Toontown Rewritten API"""

    @abstractmethod
    def news(self, *, id: Optional[int] = None, all: Optional[bool] = False) -> NewsList: 
        """Request News data from the Toontown Rewritten API"""

    @abstractmethod
    def release_notes(self, *, id: Optional[int] = None) -> ReleaseNotesList:
        """Request Release Notes data from the Toontown Rewritten API"""

    @abstractmethod
    def status(self) -> Status:
        """Request Status data from the Toontown Rewritten API"""

    @abstractmethod
    def update(self, path: Union[str, Path]) -> None:
        """Updates or creates the Toontown Rewritten files at the given Path"""


class SyncToontownClient(BaseToontownClient):
    """Synchronous client to interact with the Toontown Rewritten API"""

    def __init__(self) -> None:
        super().__init__(SyncHTTPClient())

    def connect(self) -> None:
        self.http.connect()

    def close(self) -> None:
        self.http.close()

    def status(self) -> Status:
        data = self.http.request(Route(
            'GET',
            '/status'
        ))

        return Status(**data)

    def release_notes(self, *, id: Optional[int] = None) -> ReleaseNotesList:
        if id is not None:
            path = f'/releasenotes/{id}'
        else:
            path = '/releasenotes'

        data = self.http.request(Route(
            'GET',
            path
        ))

        if not isinstance(data, list):
            data = dict(iterable=[data])
        else:
            data = dict(iterable=data)

        return ReleaseNotesList(**data)
        
    def news(self, *, id: Optional[int] = None, all: Optional[bool] = False) -> NewsList:
        if id is not None:
            path = f'/news/{id}'
        elif all is True:
            path = '/news/list'
        else:
            path = '/news'

        data = self.http.request(Route(
            'GET',
            path
        ))

        if not isinstance(data, list):
            data = dict(iterable=[data])
        else:
            data = dict(iterable=data)

        return NewsList(**data)

    def doodles(self) -> Doodles:
        data = self.http.request(Route(
            'GET',
            '/doodles'
        ))

        return Doodles(**data)

    def field_offices(self) -> FieldOffices:
        data = self.http.request(Route(
            'GET',
            '/fieldoffices'
        ))

        return FieldOffices(**data)

    def invasions(self) -> Invasions:
        data = self.http.request(Route(
            'GET',
            '/invasions'
        ))

        return Invasions(**data)

    def login(
        self, 
        *, 
        username: Optional[str] = None, 
        password: Optional[str] = None,
        app_token: Optional[str] = None,
        auth_token: Optional[str] = None,
        queue_token: Optional[str] = None,
    ) -> Login:
        params = {'format': 'json'}

        if app_token is not None and auth_token is not None:
            params['appToken'] = app_token
            params['authToken'] = auth_token
        elif queue_token is not None:
            params['queueToken'] = queue_token
        elif username is not None and password is not None:
            params['username'] = username
            params['password'] = password
        else:
            raise Exception('Please provide either a username and password, a queue token, or a auth token and app token to log in')

        data = self.http.request(Route(
            'POST',
            '/login',
            **params
        ))
        
        return Login(**data)

    def population(self) -> Population:
        data = self.http.request(Route(
            'GET',
            '/population'
        ))

        return Population(**data)

    def silly_meter(self) -> None:
        data = self.http.request(Route(
            'GET',
            '/sillymeter'
        ))

        return SillyMeter(**data)

    def update(self, path: Union[str, Path]) -> None:
        self.http.update(path)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args):
        self.close()


class AsyncToontownClient(BaseToontownClient):
    """Asynchronous client to interact with the Toontown Rewritten API"""

    def __init__(self) -> None:
        super().__init__(AsyncHTTPClient())

    async def connect(self) -> None:
        await self.http.connect()

    async def close(self) -> None:
        await self.http.close()

    async def status(self) -> Status:
        data = await self.http.request(Route(
            'GET',
            '/status'
        ))

        return Status(**data)

    async def release_notes(self, *, id: Optional[int] = None) -> ReleaseNotesList:
        if id is not None:
            path = f'/releasenotes/{id}'
        else:
            path = '/releasenotes'

        data = await self.http.request(Route(
            'GET',
            path
        ))

        if not isinstance(data, list):
            data = dict(iterable=[data])
        else:
            data = dict(iterable=data)

        return ReleaseNotesList(**data)
        
    async def news(self, *, id: Optional[int] = None, all: Optional[bool] = False) -> NewsList:
        if id is not None:
            path = f'/news/{id}'
        elif all is True:
            path = '/news/list'
        else:
            path = '/news'

        data = await self.http.request(Route(
            'GET',
            path
        ))

        if not isinstance(data, list):
            data = dict(iterable=[data])
        else:
            data = dict(iterable=data)

        return NewsList(**data)

    async def doodles(self) -> Doodles:
        data = await self.http.request(Route(
            'GET',
            '/doodles'
        ))

        return Doodles(**data)

    async def field_offices(self) -> FieldOffices:
        data = await self.http.request(Route(
            'GET',
            '/fieldoffices'
        ))

        return FieldOffices(**data)

    async def invasions(self) -> Invasions:
        data = await self.http.request(Route(
            'GET',
            '/invasions'
        ))

        return Invasions(**data)

    async def login(
        self, 
        *, 
        username: Optional[str] = None, 
        password: Optional[str] = None,
        app_token: Optional[str] = None,
        auth_token: Optional[str] = None,
        queue_token: Optional[str] = None,
    ) -> Login:
        params = {'format': 'json'}

        if app_token is not None and auth_token is not None:
            params['appToken'] = app_token
            params['authToken'] = auth_token
        elif queue_token is not None:
            params['queueToken'] = queue_token
        elif username is not None and password is not None:
            params['username'] = username
            params['password'] = password
        else:
            raise Exception('Please provide either a username and password, a queue token, or a auth token and app token to log in')

        data = await self.http.request(Route(
            'POST',
            '/login',
            **params
        ))
        
        return Login(**data)

    async def population(self) -> Population:
        data = await self.http.request(Route(
            'GET',
            '/population'
        ))

        return Population(**data)

    async def silly_meter(self) -> SillyMeter:
        data = await self.http.request(Route(
            'GET',
            '/sillymeter'
        ))

        return SillyMeter(**data)

    async def update(self, path: Union[str, Path]) -> None:
        await self.http.update(path)

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, *args):
        await self.close()



from typing import Any, Tuple

from ..exceptions import FailedResponse


__all__ = ['BaseAPIModel']


class BaseAPIModel:
    def __new__(cls, *args, **payload):
        instance = super().__new__(cls)

        if payload.get('error', None):
            """Sometimes the server will send an error field when there is no cached response"""
            raise FailedResponse(payload['error'])
            
        return instance

    def __init__(self, iterable: Tuple[Any]) -> None:
        self._iterable = iterable

    def __len__(self) -> int:
        return self._iterable.__len__()

    def __str__(self) -> str:
        return self._iterable.__str__()

    def __repr__(self) -> str:
        return self._iterable.__str__()



import asyncio
import bz2
import hashlib
import logging
import os
import platform
import sys
import time
from abc import ABC, abstractmethod
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Tuple,
    Union
)

import aiohttp
import requests
import requests.exceptions

from .exceptions import *


logger = logging.getLogger(__name__)


Session = Union[aiohttp.ClientSession, requests.Session]

BASE = 'https://www.toontownrewritten.com/api'
BASE_HEADERS = {
    'Content-Type': 'application/json',
    'User-Agent': 'toontown.py (https://github.com/jaczerob/toontown.py)'
}

LOGIN_HEADERS = {
    'Content-Type': 'application/x-www-form-urlencoded',
    'User-Agent': 'toontown.py (https://github.com/jaczerob/toontown.py)'
}

MANIFEST = 'https://cdn.toontownrewritten.com/content/patchmanifest.txt'
PATCHES = 'https://download.toontownrewritten.com/patches'

CHUNK_SIZE = 10 * 10 * 1024


def get_platform():
    if sys.platform == 'win32' and platform.machine().endswith('64'):
        return 'win64'
    return sys.platform


class Route:
    def __init__(self, method: str, path: str, **parameters: Any) -> None:
        self.method = method
        self.path = path
        self.parameters = parameters
        self.url = BASE + path

    @property
    def headers(self) -> Dict[str, Any]:
        return LOGIN_HEADERS if self.path == '/login' else BASE_HEADERS


class BaseHTTPClient(ABC):
    @abstractmethod
    def __init__(self) -> None:
        self._session: Session = None

    @abstractmethod
    def connect(self) -> None: ...

    @abstractmethod
    def close(self) -> None: ...

    @abstractmethod
    def request(self, route: Route) -> Any: ...

    @abstractmethod
    def update(self, path: Union[str, Path]) -> None: ...

    def get_outdated_files(self, manifest: Dict[str, Any], path: Path) -> List[Tuple[str, Path]]:
        files = []
        pf = get_platform()

        for file in manifest:
            if pf not in manifest[file]['only']:
                continue

            file_path: Path = path / file
            url = '{0}/{1}'.format(PATCHES, manifest[file]['dl'])

            if not file_path.exists():
                files.append((url, file_path))
                logger.info(f'Queuing {file} for download')
                continue

            hash_alg = hashlib.sha1()
            hash_alg.update(file_path.open('rb').read())

            if manifest[file]['hash'] != hash_alg.hexdigest():
                files.append((url, file_path))
                logger.info(f'Queuing {file} for download')

        return files


class SyncHTTPClient(BaseHTTPClient):
    def __init__(self) -> None:
        self._session: requests.Session = None
        self._is_closed = True

    @property
    def is_closed(self) -> bool:
        return self._is_closed

    @is_closed.setter
    def is_closed(self, value: bool) -> None:
        self._is_closed = bool(value)

    def connect(self) -> None:
        self._session = requests.Session()
        self._is_closed = False

    def close(self) -> None:
        self._session.close()
        self._is_closed = True

    def request(self, route: Route) -> Any:
        if self.is_closed:
            raise SessionNotConnected

        method = route.method
        url = route.url
        headers = route.headers
        params = route.parameters

        for tries in range(5):
            try:
                logger.info('Attempting {0} request #{1}: {2}'.format(method, tries+1, url))

                with self._session.request(method, url, params=params, headers=headers) as response:
                    response.raise_for_status()

                    data = response.json()
                    status = response.status_code

                    if 300 > status >= 200:
                        return data

                    if status in {500, 502, 504}:
                        time.sleep(1 + tries * 2)
                        continue

            except requests.exceptions.HTTPError as e:
                message = str(e)
                exc_info = type(e), e, e.__traceback__
                logger.error(message, exc_info=exc_info)

            except OSError as e:
                if tries < 4 and e.errno in {54, 10054}:
                    time.sleep(1 + tries * 2)
                    continue

                logger.info('Exhausted attempts for {0} request: {1}'.format(method, url))
                raise

    def update(self, path: Union[str, Path]) -> None:
        if isinstance(path, str):
            path = Path(path)

        if not path.is_dir():
            raise Exception('Path does not exist or is not a directory')

        manifest = self._session.get(
            MANIFEST,
            headers=BASE_HEADERS,
        ).json()

        files = self.get_outdated_files(manifest, path)

        def download(args):
            url, file_path = args
            with self._session.get(url, stream=True) as resp:
                logger.info(f'Downloading {url} to {file_path}')
                resp.raise_for_status()
                decompressor = bz2.BZ2Decompressor()

                with open(file_path, 'wb') as file:
                    for chunk in resp.iter_content(chunk_size=CHUNK_SIZE):
                        file.write(decompressor.decompress(chunk))

        with ThreadPool(os.cpu_count()) as pool:
            pool.map(download, files)


class AsyncHTTPClient(BaseHTTPClient):
    def __init__(self) -> None:
        self._session: aiohttp.ClientSession = None

    async def connect(self) -> None:
        self._session = aiohttp.ClientSession(raise_for_status=True)

    async def close(self) -> None:
        await self._session.close()

    async def request(self, route: Route) -> Any:
        if self._session.closed:
            raise SessionNotConnected

        method = route.method
        url = route.url
        headers = route.headers
        params = route.parameters

        for tries in range(5):
            try:
                logger.info('Attempting {0} request #{1}: {2}'.format(method, tries+1, url))

                async with self._session.request(method, url, params=params, headers=headers) as response:
                    data = await response.json()
                    status = response.status

                    if 300 > status >= 200:
                        return data

                    if status in {500, 502, 504}:
                        await asyncio.sleep(1 + tries * 2)
                        continue

            except aiohttp.ClientResponseError as e:
                message = str(e)
                exc_info = type(e), e, e.__traceback__
                logger.error(message, exc_info=exc_info)
                raise

            except OSError as e:
                if tries < 4 and e.errno in {54, 10054}:
                    await asyncio.sleep(1 + tries * 2)
                    continue
                raise

    async def update(self, path: Union[str, Path]) -> None:
        if isinstance(path, str):
            path = Path(path)

        if not path.is_dir():
            raise Exception('Path does not exist or is not a directory')

        manifest = await (await self._session.get(
            MANIFEST,
            headers=BASE_HEADERS,
        )).json()

        files = self.get_outdated_files(manifest, path)

        async def download(args):
            url, file_path = args

            async with self._session.get(url) as resp:
                logger.info(f'Downloading {url} to {file_path}')

                resp.raise_for_status()
                decompressor = bz2.BZ2Decompressor()

                with open(file_path, 'wb') as file:
                    async for chunk in resp.content.iter_chunked(CHUNK_SIZE):
                        file.write(decompressor.decompress(chunk))

        await asyncio.gather(*map(download, files))



from .base import *
from .doodles import *
from .fieldoffices import *
from .invasions import *
from .login import *
from .population import *
from .sillymeter import *
from .news import *
from .releasenotes import *
from .status import *



__all__ = ['Status']

class Status:
    """Wrapper class for the /status response
    
    TODO: Get rest of response (potentially)
    
    Attributes
    ----------
    open : bool
        whether or not Toontown Rewritten is open
    """
    def __init__(self, **payload) -> None:
        self.open: bool = payload.get('open', False)



from typing import Literal, Optional


__all__ = ['Login']


class Login:
    """"Wrapper class for /login response
    
    Attributes
    ----------
    success : Literal['false', 'partial', 'true', 'delayed']
        what type of response the login gave

    banner : Optional[str]
        an optional reason from the server if `false` or `partial` was given for success

    response_token : Optional[str]
        if success is `partial`, a token to verify a ToonGuard response to the server

    gameserver : Optional[str]
        if success is `true`, Toontown Rewritten's gameserver IP

    cookie : Optional[str]
        if success is `true`, your Toontown Rewritten session identifier

    eta : Optional[int]
        if success is `delayed`, how many seconds you are estimated to be able to play

    position : Optional[int]
        if success is `delayed`, how many toons are ahead of you in queue

    queue_token : Optional[str]
        if success is `delayed`, this identifier holds your position in queue
    """

    __slots__ = ['success', 'banner', 'response_token', 'gameserver', 'cookie', 'eta', 'position', 'queue_token']

    def __init__(self, **payload) -> None:
        self.success: Literal['false', 'partial', 'true', 'delayed'] = payload.get('success')
        self.banner: Optional[str] = payload.get('banner', None)
        self.response_token: Optional[str] = payload.get('responseToken', None)
        self.gameserver: Optional[str] = payload.get('gameserver', None)
        self.cookie: Optional[str] = payload.get('cookie', None)
        self.eta: Optional[int] = payload.get('eta', None)
        self.position: Optional[int] = payload.get('position', None)
        self.queue_token: Optional[str] = payload.get('queue_token', None)



from datetime import datetime
from typing import Iterator

from .base import BaseAPIModel


__all__ = ['Invasion', 'Invasions']


class Invasion:
    """Wrapper class for invasion data
    
    Attributes
    ----------
    district : str
        the district the invasion is in

    as_of : datetime
        when the invasion started

    type : str
        the Cog type of the invasion

    progress : int
        how many cogs were defeated in this invasion

    total : int
        how many cogs that are invading

    is_mega_invasion : bool
        whether or not this is a mega invasion
    """

    __slots__ = ['district', 'as_of', 'type', 'progress', 'total', 'is_mega_invasion']

    def __init__(self, district, **payload) -> None:
        self.district: str = district
        self.as_of = datetime.fromtimestamp(payload.pop('asOf'))
        self.type: str = payload.pop('type')

        progress, total = payload.pop('progress').split('/')
        self.progress = int(progress)
        self.total = int(total)
        self.is_mega_invasion: bool = total == 1000000


class Invasions(BaseAPIModel):
    """"Wrapper class for /invasions response

    A tuple-like class containing `Invasion` objects

    Attributes
    ----------
    last_updated : datetime
        the time when the invasions were last updated
    """

    __slots__ = ['last_updated']
    
    def __init__(self, **payload) -> None:
        iterable = tuple(Invasion(**{'district': key} | item) for key, item in payload.pop('invasions').items())
        super().__init__(iterable)

        self.last_updated = datetime.fromtimestamp(payload.pop('lastUpdated'))

    def __getitem__(self, index: int) -> Invasion:
        return self._iterable.__getitem__(index)

    def __iter__(self) -> Iterator[Invasion]:
        return self._iterable.__iter__()

    def __next__(self) -> Invasion:
        return next(self._iterable)
        